// -*- C++ -*-
/*!
 * @file  ImageLoader.h
 * @brief image acquisition component
 * @date  $Date$
 *
 * @author 株式会社 東日本計算センター
 *
 * $Id$
 */

#ifndef IMAGELOADER_H
#define IMAGELOADER_H

// add to code
#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/idl/InterfaceDataTypesSkel.h>
#include <iostream>
#include <fstream>
#include <string>
#include <filesystem>
#include <algorithm>
// OpenCV
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#include "local_basic_type.h"
#include "2dcdp_system.h"

// Service implementation headers
// <rtc-template block="service_impl_h">
#include "ImageLoaderControlSVC_impl.h"

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="consumer_stub_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="port_stub_h">
// </rtc-template>

using namespace RTC;

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

// add to code
/*=================================================================================================
  Macro definition
 ================================================================================================*/
#define IMLD_EXE_SUCCESS			 ( 0 )
#define IMLD_EXE_FAILURE    	 ( 1 )
#define IMLD_EXE_INIT    	     ( IMLD_EXE_FAILURE + 1 )

#define FILE_FORMAT_SUCCESS 	 ( 0 )
#define FILE_FORMAT_FAILURE 	 ( 1 )

#define IMLD_IN_INIT	       	 ( 0 )
#define IMLD_IN_PORT1       	 ( 0 )
#define IMLD_IN_PORT2        	 ( 1 )
#define IMLD_IN_PORT_MAX    	 ( IMLD_IN_PORT2 + 1 )

#define IMLD_SET_SUCCESS    	 ( 0 )
#define IMLD_SET_FAILURE    	 ( 1 )

#define IMLD_BPP_INIT		    	 ( 0 )
#define IMLD_BPP_8		 		   	 ( 8 )
#define IMLD_BPP_24			    	 ( 24)
#define IMLD_BPP_32			    	 ( 32)

#define IMLD_FILE_LOAD_DISABLE ( 0 )
#define IMLD_FILE_LOAD_ENABLE	 ( 1 )
#define IMLD_FILE_LOAD_INIT    ( 2 )
#define IMLD_FILE_LOAD_MAX     IMLD_FILE_LOAD_INIT

#define IMLD_FILE_NAME_VALID   ( 0 )
#define IMLD_FILE_NAME_INVALID ( 1 )

#define FORMAT_MARKER_MAX      ( 16)

//bitmap marker
#define FORMAT_BITMAP_B        (0x42)
#define FORMAT_BITMAP_M        (0x4D)

//jpeg marker
#define FORMAT_JPEG_1ST        (0xFF)
#define FORMAT_JPEG_2ND        (0xD8)
#define FORMAT_JPEG_3RD        (0xFF)

//png marker
#define FORMAT_PNG_1ST         (0x89)
#define FORMAT_PNG_2ND         (0x50)
#define FORMAT_PNG_3RD         (0x4E)
#define FORMAT_PNG_4TH         (0x47)

//tiff marker
#define FORMAT_TIF_1ST_TYP1    (0x49)
#define FORMAT_TIF_2ND_TYP1    (0x49)
#define FORMAT_TIF_3RD_TYP1    (0x2A)

#define FORMAT_TIF_1ST_TYP2    (0x4D)
#define FORMAT_TIF_2ND_TYP2    (0x4D)
#define FORMAT_TIF_3RD_TYP2    (0x2A)

//error message
#define EMPTY_EROR             ("ERROR:empty error of image file.")
#define UNSUPPORED_FORMAT      ("ERROR:unsupported format of image file.")
#define UNSUPPORED_BITMAP      ("ERROR:unsupported format of image file regarding bitmap.")
#define UNSUPPORED_JPEG        ("ERROR:unsupported format of image file regarding jpeg.")
#define UNSUPPORED_PNG         ("ERROR:unsupported format of image file regarding png.")
#define UNSUPPORED_TIFF        ("ERROR:unsupported format of image file regarding tiff.")
#define UNSUPPORED_EXTENSION   ("ERROR:unsupported of file extension.")
#define READING_ERROR          ("ERROR:reading of image file failed.")
#define UNSUPPORED_BPP         ("ERROR:unsupported for bit per pixel.")

#define IMLD_PATH_IN_VALID     ( 0 )
#define IMLD_PATH_IN_INVALID   ( 1 )

#define IMLD_FILE_PATH_MATCH   ( 0 )
#define IMLD_FILE_PATH_UNMATCH ( 1 )

#define FILE_PATH_OPT_ON       ( 0 )
#define FILE_PATH_OPT_OFF      ( 1 )

#define IMLD_ERROR_COUNT_MIN      ( 0 )
#define IMLD_ERROR_COUNT_RESTART  ( 10)
#define IMLD_ERROR_COUNT_MAX      ( 11)
#define IMLD_ERROR_COUNT_INIT     (255)


/*=================================================================================================
  Struct definition
 ================================================================================================*/
struct Imld_FilePath
{
  std::string path_fname1;
  std::string path_fname2;
  uWord_Type  validfname1;
  uWord_Type  validfname2;
};

/*=================================================================================================
 ================================================================================================*/

/*!
 * @class ImageLoader
 * @brief image acquisition component
 *
 * 画像ファイルのフルパス情報をデータポートから入力する
 * フルパス情報から画像データを取得し、CameraImage型に変換しデータ
 * を転送する
 * 尚、画像ファイルはbmp、jpg、png、tiffフォーマットの画像ファイル
 * を対応する
 *
 * 入力：OpenCVライブラリを使用し、画像データ(Mat型データ)を入力す
 * る
 * 出力：入力した画像データを解析し、Mat型データをCameraImage型のデ
 * ータに置き換えデータを出力する
 *
 * 1.データポートで指定された画像ファイルを解析する
 * 1-1.画像ファイルの解析
 * 　　拡張子とファイルヘッダで画像ファイルの種類（フォーマット）を
 * 判断する
 * 1-2.画像データの解析
 * 　　CopnCVのライブラリimread関数を使用し、画像の高さ/幅/bits
 * per pixelなどを判断する
 * 2.解析した画像ファイルが、bmp、jpg、png、tiffフォーマットの画像
 * であればCameraImage型のデータにデータをセットする
 *
 * OpenCVホームページ
 *
 */
class ImageLoader
  : public RTC::DataFlowComponentBase
{
 public:
  /*!
   * @brief constructor
   * @param manager Maneger Object
   */
  ImageLoader(RTC::Manager* manager);

  /*!
   * @brief destructor
   */
  ~ImageLoader();

  // <rtc-template block="public_attribute">
  
  // </rtc-template>

  // <rtc-template block="public_operation">
  
  // </rtc-template>

  /***
   * Initialization of ImageLoader RTC
   *
   * The initialize action (on CREATED->ALIVE transition)
   * formaer rtc_init_entry() 
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre 特になし
   * @post 特になし
   * 
   */
   virtual RTC::ReturnCode_t onInitialize();

  /***
   * 
   *
   * The finalize action (on ALIVE->END transition)
   * formaer rtc_exiting_entry()
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onFinalize();

  /***
   *
   * The startup action when ExecutionContext startup
   * former rtc_starting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStartup(RTC::UniqueId ec_id);

  /***
   *
   * The shutdown action when ExecutionContext stop
   * former rtc_stopping_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onShutdown(RTC::UniqueId ec_id);

  /***
   * Activate the ImageLoader RTC
   *
   * The activated action (Active state entry action)
   * former rtc_active_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre 特になし
   * @post 特になし
   * 
   */
   virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);

  /***
   * Deactivate the ImageLoader RTC
   *
   * The deactivated action (Active state exit action)
   * former rtc_active_exit()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre 特になし
   * @post 特になし
   * 
   */
   virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);

  /***
   * Periodic task of ImageLoader RTC. This task read the image
   * file.
   *
   * The execution action that is invoked periodically
   * former rtc_active_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre 画像ファイルのフルパス情報を取得するInポートが接続されて
   * いること
   * @post 特になし
   * 
   */
   virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

  /***
   * Aborting of the ImageLoader RTC
   *
   * The aborting action when main logic error occurred.
   * former rtc_aborting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre 特になし
   * @post 特になし
   * 
   */
   virtual RTC::ReturnCode_t onAborting(RTC::UniqueId ec_id);

  /***
   * Errored of the ImageLoader RTC
   *
   * The error action in ERROR state
   * former rtc_error_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre 特になし
   * @post 特になし
   * 
   */
   virtual RTC::ReturnCode_t onError(RTC::UniqueId ec_id);

  /***
   * Reset initialization of ImageLoader RTC
   *
   * The reset action that is invoked resetting
   * This is same but different the former rtc_init_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre 特になし
   * @post 特になし
   * 
   */
   virtual RTC::ReturnCode_t onReset(RTC::UniqueId ec_id);
  
  /***
   *
   * The state update action that is invoked after onExecute() action
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStateUpdate(RTC::UniqueId ec_id);

  /***
   *
   * The action that is invoked when execution context's rate is changed
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onRateChanged(RTC::UniqueId ec_id);


 protected:
  // <rtc-template block="protected_attribute">
  
  // </rtc-template>

  // <rtc-template block="protected_operation">
  
  // </rtc-template>

  // Configuration variable declaration
  // <rtc-template block="config_declare">
  /*!
   * 本Configurationパラメータで画像ファイルのパス及びファイル名を
   * 指定する
   * - Name: Imloader_file1path file1path_str
   * - DefaultValue: C:/ImageFile_Dir/ImageLoaderFile.bmp
   * - Unit: character
   * - Range: string型の為、データ範囲は特に制限はなし。推奨として
   *          256文字までとする
   * - Constraint: 指定なし（none specified）
   */
  std::string m_file1path_str;
  /*!
   * 本Configurationパラメータで画像ファイルのパス及びファイル名を
   * 指定する
   * - Name: Imloader_file2path file2path_str
   * - DefaultValue: C:/ImageFile_Dir/ImageLoaderFile.bmp
   * - Unit: character
   * - Range: string型の為、データ範囲は特に制限はなし。推奨として
   *          256文字までとする
   * - Constraint: 指定なし（none specified）
   */
  std::string m_file2path_str;
  /*!
   * 本ParameterでImloader_file1pathの入力を有効/無効設定する
   * - Name: Imloader_file1switch imld_file1sw
   * - DefaultValue: 0
   * - Unit: radio(0:disable,1:enable)
   * - Range: 0 or 1(0:disable,1:enable)
   * - Constraint: 指定なし（none specified）
   */
  int m_imld_file1sw;
  /*!
   * 本ParameterでImloader_file2pathの入力を有効/無効設定する
   * - Name: Imloader_file2switch imld_file2sw
   * - DefaultValue: 0
   * - Unit: radio(0:disable,1:enable)
   * - Range: 0 or 1(0:disable,1:enable)
   * - Constraint: 指定なし（none specified）
   */
  int m_imld_file2sw;

  // </rtc-template>

  // DataInPort declaration
  // <rtc-template block="inport_declare">
  RTC::TimedString m_in1_path;
  /*!
   * 画像ファイルが格納されているフルパス情報を取得する
   * - Type: RTC::TimedString
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: 1秒周期
   * - Operation Cycle: 1秒周期
   */
  InPort<RTC::TimedString> m_in1_pathIn;
  RTC::TimedString m_in2_path;
  /*!
   * 画像ファイルが格納されているフルパス情報を取得する
   * - Type: RTC::TimedString
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: 1秒周期
   * - Operation Cycle: 1秒周期
   */
  InPort<RTC::TimedString> m_in2_pathIn;
  
  // </rtc-template>


  // DataOutPort declaration
  // <rtc-template block="outport_declare">
  RTC::CameraImage m_out1_image;
  /*!
   * 任意のディレクトリから画像ファイルを取得し、CameraImage型に変
   * 換しデータを転送する
   * - Type: RTC::CameraImage
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: 1秒周期
   * - Operation Cycle: 1秒周期
   */
  OutPort<RTC::CameraImage> m_out1_imageOut;
  RTC::CameraImage m_out2_image;
  /*!
   * 任意のディレクトリから画像ファイルを取得し、CameraImage型に変
   * 換しデータを転送する
   * - Type: RTC::CameraImage
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: 1秒周期
   * - Operation Cycle: 1秒周期
   */
  OutPort<RTC::CameraImage> m_out2_imageOut;
  
  // </rtc-template>

  // CORBA Port declaration
  // <rtc-template block="corbaport_declare">
  /*!
   * 画像ファイル読み込み動作の制御を行う
   * Interface: 画像ファイル読み込み動作要求などのI/Fを提供する
   */
  RTC::CorbaPort m_ImageLoaderControlServPort;
  /*!
   * ImageLoader RTCのシステム制御
   * Interface: リセット制御のI/Fを提供する
   */
  RTC::CorbaPort m_ImageLoaderSystemContServPort;
  
  // </rtc-template>

  // Service declaration
  // <rtc-template block="service_declare">
  /*!
   * 画像ファイル読み込み開始/停止を要求する
   * - Argument:      画像ファイル読み込み動作のトリガ
   *                  　IMLD_TRIG_START：開始
   *                  　IMLD_TRIG_STOP：停止
   * - Return Value:  なし
   * - Exception:     なし
   * - PreCondition:  -
   * - PostCondition: -
   */
  ImageLoaderTriggerSVC_impl m_ImageLoaderTriggerProv;
  /*!
   * ImageLoader RTCのリセット制御
   * - Argument:      Resetのトリガ
   *                  　RESET_TRIG_ON：Reset On
   *                  　RESET_TRIG_OFF：Reset Off
   * - Return Value:  なし
   * - Exception:     なし
   * - PreCondition:  -
   * - PostCondition: -
   */
  ImageLoaderResetSVC_impl m_ImageLoaderResetProv;
  
  // </rtc-template>

  // Consumer declaration
  // <rtc-template block="consumer_declare">
  
  // </rtc-template>

 private:
  // <rtc-template block="private_attribute">
  
  // </rtc-template>

  // <rtc-template block="private_operation">
  
  // </rtc-template>

  // add to code
  uWord_Type nuWImageLoaderExecute( void );
  uWord_Type nuWAnalysisImageLoader( std::string, uByte_Type );
  uWord_Type nuWJudgeFileExtension( std::string, std::string * );
  uWord_Type nuWJudgeFileFormat( std::string );
  uWord_Type nuWJudgeFileFormatBitmap( uByte_Type, uByte_Type );
  uWord_Type nuWJudgeFileFormatJpeg( uByte_Type, uByte_Type, uByte_Type );
  uWord_Type nuWJudgeFileFormatPng( uByte_Type, uByte_Type, uByte_Type, uByte_Type );
  uWord_Type nuWJudgeFileFormatTiff( uByte_Type, uByte_Type, uByte_Type );
  void nvdSetImage_fDiv( Double_Type, uByte_Type );
  void nvdSetImage_Format( std::string, uByte_Type );
  void nvdSetImage_Width( const cv::Mat *, uByte_Type );
  void nvdSetImage_Height( const cv::Mat *, uByte_Type );
  void nvdSetImage_Bpp( uHWord_Type, uByte_Type );
  void nvdSetImage_Length( RTC::CameraImage *, uByte_Type );
  uWord_Type nuWSetImage_Pixels1( cv::Mat, RTC::CameraImage * );
  uWord_Type nuWSetImage_Pixels2( cv::Mat, RTC::CameraImage * );
  uWord_Type nuWGetFile1SwConfig( void );
  uWord_Type nuWGetFile2SwConfig( void );

  void nvdSetImageLoaderFilePath( void );
  void nvdJudgeImageLoaderFilePath( uWord_Type, uByte_Type );
  uWord_Type nuWGetFileSwConfig( uByte_Type );
  uWord_Type nuWJudgeNewPathIn( uByte_Type );
  uWord_Type nuWJudgeShadowPathIn( uByte_Type );
  uWord_Type nuWCompareFilePath( uByte_Type );
  void nvdSetFilePath( std::string, uByte_Type, uByte_Type );
  void nvdSetFilePathShadow( std::string, uByte_Type );
  uWord_Type nuWPathFileExists( std::string );

  // add Variables
  RTC::CameraImage m_current_image;
  uByte_Type m_nuBImldInPortNum;
  Imld_FilePath m_stFilePath;
  Imld_FilePath m_stFilePath_shadow;
  cv::Mat m_dsp_image;
  std::string m_in_path_fname;
  uWord_Type uWErrorCounter;

};


extern "C"
{
  DLL_EXPORT void ImageLoaderInit(RTC::Manager* manager);
};

#endif // IMAGELOADER_H
